/*
 * vdi_control.c - Workstation functions
 *
 * Copyright 1982 by Digital Research Inc.  All rights reserved.
 * Copyright 1999 by Caldera, Inc.
 * Copyright 2002-2020 The EmuTOS development team.
 *
 * This file is distributed under the GPL, version 2 or at your
 * option any later version.  See doc/license.txt for details.
 */

#include "emutos.h"
#include "lineavars.h"
#include "vdi_defs.h"
#include "biosbind.h"
#include "xbiosbind.h"
#include "biosext.h"
#include "asm.h"
#include "string.h"
#include "intmath.h"
#include "bdosbind.h"
#include "tosvars.h"

#define FIRST_VDI_HANDLE    1
#define LAST_VDI_HANDLE     (FIRST_VDI_HANDLE+NUM_VDI_HANDLES-1)
#define VDI_PHYS_HANDLE     FIRST_VDI_HANDLE

/*
 * entry n in the following array points to the Vwk corresponding to
 * VDI handle n.  entry 0 is unused.
 */
static Vwk *vwk_ptr[NUM_VDI_HANDLES+1];
static Vwk phys_work;          /* attribute area for physical workstation */


/*
 * template for SIZ_TAB - Returns text, line and marker sizes in device
 * coordinates. See lineavars.S for SIZE_TAB itself.
 */
static const WORD SIZ_TAB_rom[12] = {
    0,                          /* 0    min char width          */
    7,                          /* 1    min char height         */
    0,                          /* 2    max char width          */
    7,                          /* 3    max char height         */
    1,                          /* 4    min line width          */
    0,                          /* 5    reserved 0          */
    MAX_LINE_WIDTH,             /* 6    max line width          */
    0,                          /* 7    reserved 0          */
    15,                         /* 8    min marker width        */
    11,                         /* 9    min marker height       */
    120,                        /* 10   max marker width        */
    88,                         /* 11   max marker height       */
};



/* Here's the template INQ_TAB, see lineavars.S for the normal INQ_TAB */
static const WORD INQ_TAB_rom[45] = {
    1,                  /* 0  - type of alpha/graphic controllers */
    1,                  /* 1  - number of background colors  */
    0x1F,               /* 2  - text styles supported        */
    0,                  /* 3  - scale rasters = false        */
    1,                  /* 4  - number of planes         */
    0,                  /* 5  - video lookup table       */
    50,                 /* 6  - performance factor????       */
    1,                  /* 7  - contour fill capability      */
    1,                  /* 8  - character rotation capability    */
    4,                  /* 9  - number of writing modes      */
    2,                  /* 10 - highest input mode       */
    1,                  /* 11 - text alignment flag      */
    0,                  /* 12 - Inking capability        */
    0,                  /* 13 - rubber banding           */
    MAX_VERTICES,       /* 14 - maximum vertices         */
    -1,                 /* 15 - maximum intin            */
    1,                  /* 16 - number of buttons on MOUSE   */
    0,                  /* 17 - styles for wide lines            */
    0,                  /* 18 - writing modes for wide lines     */
    0,                  /* 19 - filled in with clipping flag     */

    0,                  /* 20 - extended precision pixel size information */
    0,                  /* 21 - pixel width in 1/10, 1/100 or 1/1000 microns */
    0,                  /* 22 - pixel height in 1/10, 1/100 or 1/1000 microns */
    0,                  /* 23 - horizontal resolution in dpi */
    0,                  /* 24 - vertical resolution in dpi */
    0,                  /* 25 -  */
    0,                  /* 26 -  */
    0,                  /* 27 -  */
    0,                  /* 28 - bezier flag (bit 1) */
    0,                  /* 29 -  */
    0,                  /* 30 - raster flag (bit 0), does vro_cpyfm scaling? */
    0,                  /* 31 -  */
    0,                  /* 32 -  */
    0,                  /* 33 -  */
    0,                  /* 34 -  */
    0,                  /* 35 -  */
    0,                  /* 36 -  */
    0,                  /* 37 -  */
    0,                  /* 38 -  */
    0,                  /* 39 -  */
    0,                  /* 40 - unprintable left border in pixels (printers/plotters) */
    0,                  /* 41 - unprintable upper border in pixels (printers/plotters) */
    0,                  /* 42 - unprintable right border in pixels (printers/plotters) */
    0,                  /* 43 - unprintable lower border in pixels (printers/plotters) */
    0                   /* 44 - page size (printers etc.) */
};



/* Here's the template DEV_TAB, see lineavars.S for the normal DEV_TAB! */
static const WORD DEV_TAB_rom[45] = {
    639,                        /* 0    x resolution             */
    399,                        /* 1    y resolution             */
    0,                          /* 2    device precision 0=exact,1=not exact */
    372,                        /* 3    width of pixel           */
    372,                        /* 4    height of pixel          */
    1,                          /* 5    character sizes          */
    MAX_LINE_STYLE,             /* 6    linestyles               */
    0,                          /* 7    linewidth                */
    6,                          /* 8    marker types             */
    8,                          /* 9    marker size              */
    1,                          /* 10   text font                */
    MAX_FILL_PATTERN,           /* 11   area patterns             */
    MAX_FILL_HATCH,             /* 12   crosshatch patterns       */
    2,                          /* 13   colors at one time       */
    10,                         /* 14   number of GDP's          */
    1,                          /* 15   GDP bar                  */
    2,                          /* 16   GDP arc                  */
    3,                          /* 17   GDP pie                  */
    4,                          /* 18   GDP circle               */
    5,                          /* 19   GDP ellipse              */
    6,                          /* 20   GDP elliptical arc       */
    7,                          /* 21   GDP elliptical pie       */
    8,                          /* 22   GDP rounded rectangle    */
    9,                          /* 23   GDP filled rounded rectangle */
    10,                         /* 24   GDP #justified text      */
    3,                          /* 25   GDP #1                   */
    0,                          /* 26   GDP #2                   */
    3,                          /* 27   GDP #3                   */
    3,                          /* 28   GDP #4                   */
    3,                          /* 29   GDP #5                   */
    0,                          /* 30   GDP #6                   */
    3,                          /* 31   GDP #7                   */
    0,                          /* 32   GDP #8                   */
    3,                          /* 33   GDP #9                   */
    2,                          /* 34   GDP #10                  */
    0,                          /* 35   Color capability         */
    1,                          /* 36   Text Rotation            */
    1,                          /* 37   Polygonfill              */
    0,                          /* 38   Cell Array               */
    2,                          /* 39   Palette size             */
    2,                          /* 40   # of locator devices 1 = mouse */
    1,                          /* 41   # of valuator devices    */
    1,                          /* 42   # of choice devices      */
    1,                          /* 43   # of string devices      */
    2                           /* 44   Workstation Type 2 = out/in */
};



Vwk * get_vwk_by_handle(WORD handle)
{
    if ((handle < FIRST_VDI_HANDLE) || (handle > LAST_VDI_HANDLE))
        return NULL;

    return vwk_ptr[handle];
}



/*
 * validate colour index
 *
 * checks the supplied colour index and, if valid, returns it;
 * otherwise returns 1 (which by default maps to black)
 */
WORD validate_color_index(WORD colnum)
{
    if ((colnum < 0) || (colnum >= numcolors))
        return 1;

    return colnum;
}



/* Set Clip Region */
void vdi_vs_clip(Vwk * vwk)
{
    vwk->clip = INTIN[0];
    if (vwk->clip) {
        Rect * rect = (Rect*)PTSIN;
        arb_corner(rect);
        vwk->xmn_clip = max(0, rect->x1);
        vwk->ymn_clip = max(0, rect->y1);
        vwk->xmx_clip = min(xres, rect->x2);
        vwk->ymx_clip = min(yres, rect->y2);
    } else {
        vwk->xmn_clip = 0;
        vwk->ymn_clip = 0;
        vwk->xmx_clip = xres;
        vwk->ymx_clip = yres;
    }
}



/* SET_WRITING_MODE: */
void vdi_vswr_mode(Vwk * vwk)
{
    WORD wm;

    CONTRL[4] = 1;
    wm = ((INTIN[0]<MIN_WRT_MODE) || (INTIN[0]>MAX_WRT_MODE)) ? DEF_WRT_MODE : INTIN[0];

    INTOUT[0] = wm;
    vwk->wrt_mode = wm - 1;
}



static void init_wk(Vwk * vwk)
{
    WORD l;
    WORD *pointer;
    const WORD *src_ptr;

    pointer = INTIN;
    pointer++;

    l = *pointer++;             /* INTIN[1] */
    if ((l > MAX_LINE_STYLE) || (l < MIN_LINE_STYLE))
        l = DEF_LINE_STYLE;
    vwk->line_index = l - 1;

    l = validate_color_index(*pointer++);   /* INTIN[2] */
    vwk->line_color = MAP_COL[l];

    l = *pointer++;             /* INTIN[3] */
    if ((l > MAX_MARK_STYLE) || (l < MIN_MARK_STYLE))
        l = DEF_MARK_STYLE;
    vwk->mark_index = l - 1;

    l = validate_color_index(*pointer++);   /* INTIN[4] */
    vwk->mark_color = MAP_COL[l];

    /* You always get the default font */
    pointer++;                  /* INTIN[5] */

    l = validate_color_index(*pointer++);   /* INTIN[6] */
    vwk->text_color = MAP_COL[l];

    vwk->mark_height = DEF_MKHT;
    vwk->mark_scale = 1;

    l = *pointer++;             /* INTIN[7] */
    vwk->fill_style = ((l > MAX_FILL_STYLE) || (l < MIN_FILL_STYLE)) ? DEF_FILL_STYLE : l;

    l = *pointer++;             /* INTIN[8] */
    if (vwk->fill_style == FIS_PATTERN)
        l = ((l > MAX_FILL_PATTERN) || (l < MIN_FILL_PATTERN)) ? DEF_FILL_PATTERN : l;
    else
        l = ((l > MAX_FILL_HATCH) || (l < MIN_FILL_HATCH)) ? DEF_FILL_HATCH : l;
    vwk->fill_index = l;

    l = validate_color_index(*pointer++);   /* INTIN[9] */
    vwk->fill_color = MAP_COL[l];

    vwk->xfm_mode = *pointer;      /* INTIN[10] */

    st_fl_ptr(vwk);                /* set the fill pattern as requested */

    vwk->wrt_mode = WM_REPLACE;    /* default is replace mode */
    vwk->line_width = DEF_LWID;
    vwk->line_beg = SQUARED;       /* default to squared ends */
    vwk->line_end = SQUARED;

    vwk->fill_per = TRUE;

    vwk->xmn_clip = 0;
    vwk->ymn_clip = 0;
    vwk->xmx_clip = xres;
    vwk->ymx_clip = yres;
    vwk->clip = FALSE;

    text_init2(vwk);

    /* move default user defined pattern to RAM */
    pointer = &vwk->ud_patrn[0];
    src_ptr = (const WORD *)ROM_UD_PATRN;
    for (l = 0; l < 16; l++)
        *pointer++ = *src_ptr++;

    vwk->multifill = 0;
    vwk->ud_ls = LINE_STYLE[0];

    CONTRL[2] = 6;
    CONTRL[4] = 45;

    pointer = INTOUT;
    src_ptr = DEV_TAB;
    for (l = 0; l < 45; l++)
        *pointer++ = *src_ptr++;

    pointer = PTSOUT;
    src_ptr = SIZ_TAB;
    for (l = 0; l < 12; l++)
        *pointer++ = *src_ptr++;

    /* setup initial bezier values */
    vwk->bez_qual = 7;
#if 0
    vwk->bezier.available = 1;
    vwk->bezier.depth_scale.min = 9;
    vwk->bezier.depth_scale.max = 0;
    vwk->bezier.depth.min = 2;
    vwk->bezier.depth.max = 7;
#endif

    vwk->next_work = NULL;  /* neatness */

    flip_y = 1;
}



void vdi_v_opnvwk(Vwk * vwk)
{
    WORD handle;
    Vwk **p;

    /*
     * ensure that CUR_WORK always points to a valid workstation
     * even if v_opnvwk() exits early.
     */
    CUR_WORK = &phys_work;

    /* First find a free handle */
    for (handle = VDI_PHYS_HANDLE+1, p = vwk_ptr+handle; handle <= LAST_VDI_HANDLE; handle++, p++) {
        if (!*p) {
            *p = vwk;
            break;
        }
    }
    if (handle > LAST_VDI_HANDLE) { /* No handle available, exit */
        CONTRL[6] = 0;
        return;
    }

    /*
     * Allocate the memory for a virtual workstation
     */
    vwk = (Vwk *)Malloc(sizeof(Vwk));
    if (vwk == NULL) {
        CONTRL[6] = 0;  /* No memory available, exit */
        return;
    }

    vwk_ptr[handle] = vwk;
    vwk->handle = CONTRL[6] = handle;
    init_wk(vwk);
    CUR_WORK = vwk;
}

void vdi_v_clsvwk(Vwk * vwk)
{
    WORD handle;

    /* vwk points to workstation to deallocate */
    handle = vwk->handle;
    if (handle == VDI_PHYS_HANDLE)  /* can't close physical this way */
        return;
    if (!vwk_ptr[handle])           /* workstation is already closed */
        return;

    vwk_ptr[handle] = NULL;         /* close it */

    /*
     * When we close a virtual workstation, Atari TOS and previous versions
     * of EmuTOS update CUR_WORK (line-A's idea of the current workstation)
     * to point to the precursor of the closed workstation.  This is a bit
     * arbitrary, especially as the workstation being closed isn't necessarily
     * what line-A thinks is the current one.
     *
     * What we must do as a minimum is ensure that CUR_WORK points to a
     * valid open workstation.  The following does that by pointing it to
     * the physical workstation.  That's what NVDI appears to do too.
     */
    CUR_WORK = &phys_work;

    Mfree(vwk);
}



/* OPEN_WORKSTATION: */
void vdi_v_opnwk(Vwk * vwk)
{
    int i;
    Vwk **p;
    WORD newrez;

    /*
     * Programs can request a video mode switch by passing the desired
     * mode + 2 in INTIN[0]. Falcon-specific modes are currently not
     * supported.
     */
    newrez = INTIN[0] - 2;
    if (
        (newrez == ST_LOW) || (newrez == ST_MEDIUM) || (newrez == ST_HIGH)
#if CONF_WITH_TT_SHIFTER
        || (newrez == TT_LOW) || (newrez == TT_MEDIUM)
#endif
       ) {
        if (newrez != Getrez()) {
            Setscreen(-1L, -1L, newrez, 0);
        }
    }

    /* We need to copy some initial table data from the ROM */
    for (i = 0; i < 12; i++) {
        SIZ_TAB[i] = SIZ_TAB_rom[i];
    }

    for (i = 0; i < 45; i++) {
        DEV_TAB[i] = DEV_TAB_rom[i];
        INQ_TAB[i] = INQ_TAB_rom[i];
    }

    /* Copy data from line-A variables */
    xres = V_REZ_HZ - 1;        /* xres/yres are DEV_TAB[0]/[1] */
    yres = V_REZ_VT - 1;
    INQ_TAB[4] = v_planes;

    /* get pixel sizes for use by routines in vdi_gdp.c & vdi_line.c */
    get_pixel_size(&xsize,&ysize);  /* xsize/ysize are DEV_TAB[3]/[4] */

    /* Indicate whether LUT is supported */
    if ((INQ_TAB[4] == 16) || (get_monitor_type() == MON_MONO))
        INQ_TAB[5] = 0;
    else INQ_TAB[5] = 1;

    /* Calculate colors allowed at one time */
    if (INQ_TAB[4] < 8)
        numcolors = 2<<(v_planes-1);/* numcolors is DEV_TAB[13] */
    else
        numcolors = 256;

    /* get palette (COLOR.CPX from the MegaSTe language disk cares about this) */
    DEV_TAB[39] = get_palette();

    /* initialize the vwk pointer array */
    vwk = &phys_work;
    vwk_ptr[VDI_PHYS_HANDLE] = vwk;
    CONTRL[6] = vwk->handle = VDI_PHYS_HANDLE;
    for (i = VDI_PHYS_HANDLE+1, p = vwk_ptr+i; i <= LAST_VDI_HANDLE; i++)
        *p++ = NULL;

    line_cw = -1;               /* invalidate current line width */

    init_colors();              /* Initialize palette etc. */

    text_init();                /* initialize the SIZ_TAB info */

    init_wk(vwk);

    timer_init();
    vdimouse_init();            /* initialize mouse */
    esc_init(vwk);              /* enter graphics mode */

    /* Just like TOS 2.06, make the physical workstation the current workstation for Line-A. */
    CUR_WORK = vwk;
}



/* CLOSE_WORKSTATION: */
void vdi_v_clswk(Vwk * vwk)
{
    WORD handle;
    Vwk **p;

    /* close all open virtual workstations */
    for (handle = VDI_PHYS_HANDLE+1, p = vwk_ptr+handle; handle <= LAST_VDI_HANDLE; handle++, p++) {
        if (*p) {
            Mfree(*p);
            *p = NULL;
        }
    }
    CUR_WORK = vwk_ptr[VDI_PHYS_HANDLE];

    timer_exit();
    vdimouse_exit();                    /* deinitialize mouse */
    esc_exit(vwk);                      /* back to console mode */
}



/*
 * vdi_v_clrwk - clear screen
 *
 * Screen is cleared from the base address v_bas_ad.
 */
void vdi_v_clrwk(Vwk * vwk)
{
    ULONG size;

    /* Calculate screen size */
    size = (ULONG)v_lin_wr * V_REZ_VT;

    /* clear the screen */
    bzero(v_bas_ad, size);
}



/*
 * vdi_vq_extnd - Extended workstation inquire
 */
void vdi_vq_extnd(Vwk * vwk)
{
    WORD i;
    WORD *dst, *src;

    CONTRL[2] = 6;
    CONTRL[4] = 45;

    flip_y = 1;
    dst = PTSOUT;
    if (*(INTIN) == 0) {
        src = SIZ_TAB;
        for (i = 0; i < 12; i++)
            *dst++ = *src++;

        src = DEV_TAB;
    }
    else {
        /* copy the clipping ranges to PTSOUT */
        *dst++ = vwk->xmn_clip;       /* PTSOUT[0] */
        *dst++ = vwk->ymn_clip;       /* PTSOUT[1] */
        *dst++ = vwk->xmx_clip;       /* PTSOUT[2] */
        *dst++ = vwk->ymx_clip;       /* PTSOUT[3] */

        for (i = 4; i < 12; i++)
            *dst++ = 0;

        src = INQ_TAB;
        INQ_TAB[19] = vwk->clip;      /* now update INQTAB */
    }

    /* copy DEV_TAB or INQ_TAB to INTOUT */
    dst = INTOUT;
    for (i = 0; i < 45; i++)
        *dst++ = *src++;
}
